## Advanced OOP
C++ program that demonstrates virtual functions, polymorphism, abstract classes, 
and dynamic casting in the context of differential equation solvers and computational 
fluid dynamics simulations. Here's the implementation:

### Resources [(C++ Tutorial)](https://www.geeksforgeeks.org/c-plus-plus/?ref=shm)

[algorithm](../html_includes/algorithm.html),
[auto](../html_src/auto.html),
[class](../html_src/class.html),
[command_line_arguments](../html_src/command_line_arguments.html),
[destructor](../html_src/destructor.html),
[constructor](../html_src/constructor.html),
[dynamic_cast](../html_src/dynamic_cast.html),
[iostream](../html_includes/iostream.html),
[memory](../html_includes/memory.html),
[namespace](../html_src/namespace.html),
[override](../html_src/override.html),
[range-based](../html_src/range-based.html),
[std::cin](../html_includes/iostream.html),
[std::cout](../html_includes/iostream.html),
[std::cerr](../html_includes/iostream.html),
[subclass](../html_src/subclass.html),
[vector](../html_includes/vector.html),
[virtual](../html_includes/virtual_function.html),
[void](../html_includes/void.html)


```cpp
#include <iostream>
#include <vector>
#include <memory>

// Abstract base class for differential equation solvers
class DifferentialEquationSolver {
public:
    virtual ~DifferentialEquationSolver() = default;
    virtual void solve() = 0; // Pure virtual function
    virtual void initialize() { std::cout << "Initializing solver\n"; }
};

// Derived class for Euler method
class EulerSolver : public DifferentialEquationSolver {
public:
    void solve() override {
        std::cout << "Solving using Euler method\n";
    }
};

// Derived class for Runge-Kutta method
class RungeKuttaSolver : public DifferentialEquationSolver {
public:
    void solve() override {
        std::cout << "Solving using Runge-Kutta method\n";
    }
};

// Abstract base class for boundary conditions
class BoundaryCondition {
public:
    virtual ~BoundaryCondition() = default;
    virtual void apply() = 0; // Pure virtual function
};

// Derived class for Dirichlet boundary condition
class DirichletBoundaryCondition : public BoundaryCondition {
public:
    void apply() override {
        std::cout << "Applying Dirichlet boundary condition\n";
    }
};

// Derived class for Neumann boundary condition
class NeumannBoundaryCondition : public BoundaryCondition {
public:
    void apply() override {
        std::cout << "Applying Neumann boundary condition\n";
    }
};

// Base Solver class with adaptive time-stepping capability
class Solver {
public:
    virtual ~Solver() = default;
    virtual void solve() = 0;
};

// Derived AdaptiveSolver class
class AdaptiveSolver : public Solver {
public:
    void solve() override {
        std::cout << "Solving with adaptive time-stepping\n";
    }

    void adjustTimeStep() {
        std::cout << "Adjusting time step\n";
    }
};

// Function to demonstrate polymorphism
void solveProblem(DifferentialEquationSolver& solver) {
    solver.initialize();
    solver.solve();
}

int main() {
    // Demonstrate virtual functions and polymorphism
    std::cout << "Demonstrating virtual functions and polymorphism:\n";
    EulerSolver eulerSolver;
    RungeKuttaSolver rkSolver;

    solveProblem(eulerSolver);
    solveProblem(rkSolver);

    // Demonstrate abstract class usage
    std::cout << "\nDemonstrating abstract class usage:\n";
    std::vector<std::unique_ptr<BoundaryCondition>> boundaryConditions;
    boundaryConditions.push_back(std::make_unique<DirichletBoundaryCondition>());
    boundaryConditions.push_back(std::make_unique<NeumannBoundaryCondition>());

    for (const auto& bc : boundaryConditions) {
        bc->apply();
    }

    // Demonstrate dynamic_cast
    std::cout << "\nDemonstrating dynamic_cast:\n";
    std::vector<std::unique_ptr<Solver>> solvers;
    solvers.push_back(std::make_unique<AdaptiveSolver>());
    solvers.push_back(std::make_unique<Solver>());

    for (const auto& solver : solvers) {
        solver->solve();
        
        // Try to cast to AdaptiveSolver
        if (auto adaptiveSolver = dynamic_cast<AdaptiveSolver*>(solver.get())) {
            std::cout << "This is an AdaptiveSolver. ";
            adaptiveSolver->adjustTimeStep();
        } else {
            std::cout << "This is not an AdaptiveSolver.\n";
        }
    }

    return 0;
}
```

This code demonstrates the requested concepts:

1. **Virtual Functions and Polymorphism**: 
   - The `DifferentialEquationSolver` base class has a virtual destructor and a pure virtual `solve()` function.
   - `EulerSolver` and `RungeKuttaSolver` derive from `DifferentialEquationSolver` and override the `solve()` function.
   - The `solveProblem()` function demonstrates polymorphism by accepting any `DifferentialEquationSolver` object.

2. **Abstract Class**:
   - `BoundaryCondition` is an abstract base class with a pure virtual `apply()` function.
   - `DirichletBoundaryCondition` and `NeumannBoundaryCondition` are concrete implementations of `BoundaryCondition`.

3. **Dynamic Cast**:
   - The `Solver` base class and `AdaptiveSolver` derived class are used to demonstrate `dynamic_cast`.
   - We use `dynamic_cast` to safely check if a `Solver` pointer can be cast to an `AdaptiveSolver` pointer, allowing us to call the `adjustTimeStep()` method only on `AdaptiveSolver` instances.

This code provides a framework for implementing different numerical methods for solving differential equations, handling various boundary conditions in CFD simulations, and safely working with a hierarchy of solver types. It demonstrates key C++ features useful in scientific computing and numerical simulations.

## Related 
- How can I implement polymorphism in C++ for different numerical methods
- What are the best practices for defining abstract classes in C++
- How can I safely use dynamic_cast in a class hierarchy
- What are some common pitfalls when implementing pure virtual functions
- How can I ensure correct memory management when using derived classes
