#include <iostream>
#include <vector>
#include <memory>

// Abstract base class for differential equation solvers
class DifferentialEquationSolver {
public:
    virtual ~DifferentialEquationSolver() = default;
    virtual void solve() = 0; // Pure virtual function
    virtual void initialize() { std::cout << "Initializing solver\n"; }
};

// Derived class for Euler method
class EulerSolver : public DifferentialEquationSolver {
public:
    void solve() override {
        std::cout << "Solving using Euler method\n";
    }
};

// Derived class for Runge-Kutta method
class RungeKuttaSolver : public DifferentialEquationSolver {
public:
    void solve() override {
        std::cout << "Solving using Runge-Kutta method\n";
    }
};

// Abstract base class for boundary conditions
class BoundaryCondition {
public:
    virtual ~BoundaryCondition() = default;
    virtual void apply() = 0; // Pure virtual function
};

// Derived class for Dirichlet boundary condition
class DirichletBoundaryCondition : public BoundaryCondition {
public:
    void apply() override {
        std::cout << "Applying Dirichlet boundary condition\n";
    }
};

// Derived class for Neumann boundary condition
class NeumannBoundaryCondition : public BoundaryCondition {
public:
    void apply() override {
        std::cout << "Applying Neumann boundary condition\n";
    }
};

// Base Solver class with adaptive time-stepping capability
class Solver {
public:
    virtual ~Solver() = default;
    virtual void solve() = 0;
};

// Derived AdaptiveSolver class
class AdaptiveSolver : public Solver {
public:
    void solve() override {
        std::cout << "Solving with adaptive time-stepping\n";
    }

    void adjustTimeStep() {
        std::cout << "Adjusting time step\n";
    }
};

// Function to demonstrate polymorphism
void solveProblem(DifferentialEquationSolver& solver) {
    solver.initialize();
    solver.solve();
}


int main() {
    // Demonstrate virtual functions and polymorphism
    std::cout << "Demonstrating virtual functions and polymorphism:\n";
    EulerSolver eulerSolver;
    RungeKuttaSolver rkSolver;

    solveProblem(eulerSolver);
    solveProblem(rkSolver);

    // Demonstrate abstract class usage
    std::cout << "\nDemonstrating abstract class usage:\n";
    std::vector<std::unique_ptr<BoundaryCondition>> boundaryConditions;
    boundaryConditions.push_back(std::make_unique<DirichletBoundaryCondition>());
    boundaryConditions.push_back(std::make_unique<NeumannBoundaryCondition>());

    for (const auto& bc : boundaryConditions) {
        bc->apply();
    }

    // Demonstrate dynamic_cast
    std::cout << "\nDemonstrating dynamic_cast:\n";
    std::vector<std::unique_ptr<Solver>> solvers;
    solvers.push_back(std::make_unique<AdaptiveSolver>());
    // Remove the following line since Solver is abstract:
    // solvers.push_back(std::make_unique<Solver>());

    for (const auto& solver : solvers) {
        solver->solve();

        // Try to cast to AdaptiveSolver
        if (auto adaptiveSolver = dynamic_cast<AdaptiveSolver*>(solver.get())) {
            std::cout << "This is an AdaptiveSolver. ";
            adaptiveSolver->adjustTimeStep();
        } else {
            std::cout << "This is not an AdaptiveSolver.\n";
        }
    }

    return 0;
}

